"""Static and environment variables."""
from os import getenv

WSGI_HOST = getenv('WSGI_HOST', '127.0.0.1')
WSGI_PORT = getenv('WSGI_PORT', '5000')
WSGI_WORKERS = getenv('WSGI_WORKERS', '1')
WSGI_OPTIONS = {
    'bind': '{}:{}'.format(WSGI_HOST, WSGI_PORT),
    'workers': int(WSGI_WORKERS),
}



"""Base class for serving predictions."""
from flask import Flask, request
from flask_restful import Resource, Api

from .config import WSGI_OPTIONS
from .wsgi import StandaloneApplication
from .utils import make_serializable


class PredictionServer(object):
    """Easy deploy class."""

    def __init__(self, predict):
        """Initialize class with prediction function."""
        self.predict = predict
        self.app = Flask('{}_{}'.format(self.__class__.__name__, type(predict).__name__))
        self.api = Api(self.app)
        self._create_prediction_endpoint()

    def __repr__(self):
        """String representation."""
        return '<PredictionsServer: {}>'.format(type(self.predict).__name__)

    def _create_prediction_endpoint(self):
        """Create an endpoint to serve predictions."""
        # copy instance variables to local scope for resource class
        predict = self.predict
        logger = self.app.logger

        # create restful resource
        class Predictions(Resource):
            def post(self):
                data = request.get_json()
                logger.debug(data)
                prediction = predict(data)
                logger.debug(prediction)
                return prediction.tolist()

        # map resource to endpoint
        self.api.add_resource(Predictions, '/predictions')

    def create_info_endpoint(self, name, data):
        """Create an endpoint to serve info GET requests."""
        # make sure data is serializable
        data = make_serializable(data)

        # create generic restful resource to serve static JSON data
        class InfoBase(Resource):
            def get(self):
                return data

        def info_factory(name):
            """Return an Info derivative resource."""
            class NewClass(InfoBase):
                pass
            NewClass.__name__ = "{}_{}".format(name, InfoBase.__name__)
            return NewClass

        self.api.add_resource(info_factory(name), '/info/{}'.format(name))

    def serve(self):
        """Serve predictions as an API endpoint."""
        self.server = StandaloneApplication(self.app, WSGI_OPTIONS).run()



"""WSGI server implementation."""
from __future__ import unicode_literals
import multiprocessing
import gunicorn.app.base
from gunicorn.six import iteritems


def number_of_workers():
    """Return the number of workers for the server to use."""
    return (multiprocessing.cpu_count() * 2) - 1


def handler_app(environ, start_response):
    """Placeholder app."""
    response_body = b'Works fine'
    status = '200 OK'

    response_headers = [
        ('Content-Type', 'text/plain'),
    ]

    start_response(status, response_headers)

    return [response_body]


class StandaloneApplication(gunicorn.app.base.BaseApplication):
    """Standalone gunicorn server."""

    def __init__(self, app, options=None):
        """Initialize server."""
        self.options = options or {}
        self.application = app
        super(StandaloneApplication, self).__init__()

    def load_config(self):
        """Load server configuration."""
        config = dict([(key, value) for key, value in iteritems(self.options)
                       if key in self.cfg.settings and value is not None])
        for key, value in iteritems(config):
            self.cfg.set(key.lower(), value)

    def load(self):
        """Return the underlying application."""
        return self.application


if __name__ == '__main__':
    options = {
        'bind': '%s:%s' % ('127.0.0.1', '8080'),
        'workers': number_of_workers(),
    }
    StandaloneApplication(handler_app, options).run()
